import pytest
from ophydregistry.exceptions import ComponentNotFound

from haven.instrument import Robot, load_robots


def test_robot():
    robot = Robot("25idAustin", name="robot")
    # Check PVs are correct
    assert robot.i.user_readback.pvname == "25idAustin:i.RBV"
    assert robot.samples.sample8.present.pvname == "25idAustin:sample8:present"
    assert robot.samples.sample8.rz.pvname == "25idAustin:sample8:rz"


def test_load_robot(sim_registry):
    load_robots()
    # Test the robot info is extracted properly
    rbt = sim_registry.find(label="robots")
    assert rbt.name == "A"
    assert rbt.prefix == "255idAustin"


def test_load_no_robot(sim_registry):
    load_robots(config={})

    # Test the robot is not in config
    result = pytest.raises(ComponentNotFound, sim_registry.findall, label="robots")

    # Assert that the expected exception is raised
    assert 'Could not find components matching: label="robots", name="None"' in str(
        result.value
    )


# -----------------------------------------------------------------------------
# :author:    Yanna Chen
# :email:     yannachen@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
